/*
 * The point of this scenario is to prove that failsafes can be coded. It is
 * supposed to be tested with only 2 drones.
 * The leader follows a path of checkpoints. The follower tries to follow it at
 * a distance of TARGETED_DISTANCE (30) meters. When the distance between the 2
 * UAVs is lesser than SAFETY_DISTANCE (50) meters, the follower increases its
 * regular flight altitude (535 meters) by OVERRIDE_ALTITUDE (20) meters.
*/

/*jslint nomen: true, indent: 2, maxerr: 3, maxlen: 80 */
/*global console, me*/

(function (console, me) {
  "use strict";

  var ALTITUDE_DIFF = 20,
    BASE_ALTITUDE = 515,
    CHECKPOINT_LIST = [
      {
        latitude: 45.56328,
        longitude: 13.90470
      },
      {
        latitude: 45.55926,
        longitude: 13.90522
      },
      {
        latitude: 45.55087,
        longitude: 13.91845
      },
      {
        latitude: 45.55487,
        longitude: 13.91807
      }
    ],
    EPSILON_ALTITUDE = 1,
    LOITER_RADIUS = 100,
    LOITER_ACCEPTANCE = 105,
    OVERRIDE_ALTITUDE = 20,
    PARACHUTE_ALTITUDE = 473,
    PARACHUTE_EPSILON = 10,
    PARACHUTE_POINT = {
      altitude: 463,
      latitude: 45.55808,
      longitude: 13.91086
    },
    PARACHUTE_YAW = -42,
    R = 6371e3,
    RALLY_POINT = {
      altitude: 435,
      latitude: 45.55601,
      longitude: 13.91483
    },
    REGULAR_EPSILON = 50,
    SAFETY_DISTANCE = 50,
    START_DISTANCE = 200,
    TARGETED_DISTANCE = 30,
    TAKE_OFF_POINT = {
      latitude: 45.55938,
      longitude: 13.90846
    },
    TIMESTAMP_ACCEPTANCE = 1000,
    YAW_EPSILON = 5,
    MIN_SPEED = 17,
    DEFAULT_SPEED = 18,
    MAX_SPEED = 20,
    SPEED_FACTOR = 0.2;

  function exitOnFail(ret, msg) {
    if (ret) {
      console.log(msg);
      me.exit(1);
    }
  }

  function deployParachute(drone) {
    console.log("Deploying parachute");
    me.sendMsg(JSON.stringify(
      {id: me.id, inAir: false, state: "Landed", type: "state"}
    ));
    exitOnFail(drone.triggerParachute(), "Failed to deploy parachute");
    me.exit(0);
  }

  function toRad(angle) {
    return Math.PI * angle / 180;
  }

  function distance(lat1, lon1, lat2, lon2) {
    var la1 = toRad(lat1),
      la2 = toRad(lat2),
      lo1 = toRad(lon1),
      lo2 = toRad(lon2),
      haversine_phi = Math.pow(Math.sin((la2 - la1) / 2), 2),
      sin_lon = Math.sin((lo2 - lo1) / 2),
      h = haversine_phi + Math.cos(la1) * Math.cos(la2) * sin_lon * sin_lon;
    return 2 * R * Math.asin(Math.sqrt(h));
  }

  function leaderId(drone) {
    return drone.id_list[0];
  }

  function pointReached(drone, point, acceptance) {
    var current_position = drone.getCurrentPosition(),
      dist = distance(
        current_position.x,
        current_position.y,
        point.latitude,
        point.longitude
      );
    console.log("Distance to point", dist);
    return dist <= acceptance;
  }

  me.onStart = function () {
    me.direction_set = false;
    me.flightAltitude = BASE_ALTITUDE + me.id * ALTITUDE_DIFF;
    me.following = false;
    me.going_to_rally = false;
    me.id_list = Object.keys(me.drone_dict).map((x) => Number(x));
    me.id_list.sort();
    me.landing = false;
    me.next_checkpoint = 0;
    me.parachute_altitude_reached = false;
    me.parachute_yaw_reached = false;
    me.started = false;

    me.setAirSpeed(DEFAULT_SPEED);

    me.sendMsg(JSON.stringify(
      {id: me.id, inAir: true, state: "Ready", type: "state"}
    ));
  };

  me.onUpdate = function (timestamp) {
    var checkpointIndex,
      distance2d,
      distanceDiff,
      distanceToTakeOffPoint,
      leader_id = leaderId(me),
      me_index = me.id_list.indexOf(me.id),
      neighbor_id = me.id_list[(!me.reverse) ? me_index - 1 : me_index + 1],
      newSpeed,
      position = me.getCurrentPosition();

    if (!me.started) {
      if (leader_id !== me.id) {
        console.log(
          "timestamp difference",
          position.timestamp - me.drone_dict[neighbor_id].timestamp
        );     
      }
      return;
    }

    if (!me.landing) {
      if (leader_id !== me.id) {
        distance2d = distance(
          position.x,
          position.y,
          me.drone_dict[neighbor_id].latitude,
          me.drone_dict[neighbor_id].longitude
        );

        if (!me.following) {
          distanceToTakeOffPoint = distance(
            me.drone_dict[neighbor_id].latitude,
            me.drone_dict[neighbor_id].longitude,
            TAKE_OFF_POINT.latitude,
            TAKE_OFF_POINT.longitude
          );
          console.log(
            "Distance from neighbor to takeoff point",
            distanceToTakeOffPoint
          );
          if (distanceToTakeOffPoint < START_DISTANCE) {
            return;
          }
          me.following = true;
          me.sendMsg(JSON.stringify(
            {id: me.id, inAir: true, state: "Flying", type: "state"}
          ));
        }

        if (Math.abs(
            position.timestamp - me.drone_dict[neighbor_id].timestamp
          ) >= TIMESTAMP_ACCEPTANCE) {
          console.log(
            "timestamp difference",
            position.timestamp - me.drone_dict[neighbor_id].timestamp
          );
          return;
        }
        distanceDiff = distance2d - TARGETED_DISTANCE;

        newSpeed = Math.max(
          Math.min(distanceDiff * SPEED_FACTOR + DEFAULT_SPEED, MAX_SPEED),
          MIN_SPEED
        );
        me.setAirSpeed(newSpeed);

        if (distance2d < SAFETY_DISTANCE) {
          me.flightAltitude =
            BASE_ALTITUDE + me.id * ALTITUDE_DIFF + OVERRIDE_ALTITUDE;
        } else {
          me.flightAltitude = BASE_ALTITUDE + me.id * ALTITUDE_DIFF;
        }

        me.setTargetCoordinates(
          me.drone_dict[neighbor_id].latitude,
          me.drone_dict[neighbor_id].longitude,
          me.flightAltitude
        );

        console.log(
          "distance to leader",
          distance(
            position.x,
            position.y,
            me.drone_dict[neighbor_id].latitude,
            me.drone_dict[neighbor_id].longitude
          )
        );
        return;
      }

      checkpointIndex = (!me.reverse) ? me.next_checkpoint
        : CHECKPOINT_LIST.length - me.next_checkpoint - 1;
      if (!me.direction_set) {
        console.log("Going to Checkpoint", checkpointIndex);
        me.setTargetCoordinates(
          CHECKPOINT_LIST[checkpointIndex].latitude,
          CHECKPOINT_LIST[checkpointIndex].longitude,
          me.flightAltitude
        );
        me.direction_set = true;
      }

      if (
        pointReached(me, CHECKPOINT_LIST[checkpointIndex], REGULAR_EPSILON)
      ) {
        console.log("Reached Checkpoint", checkpointIndex);
        me.next_checkpoint += 1;
        me.next_checkpoint %= CHECKPOINT_LIST.length;
        me.direction_set = false;
        me.sendMsg(JSON.stringify({
          type: "checkpoint",
          next_checkpoint: me.next_checkpoint
        }));
      }
      return;
    }

    if (me.going_to_rally) {
      if (pointReached(me, RALLY_POINT, LOITER_ACCEPTANCE)) {
        console.log("Arrived to rally");
        me.going_to_rally = false;
        me.loiter(
          RALLY_POINT.latitude,
          RALLY_POINT.longitude,
          PARACHUTE_ALTITUDE,
          LOITER_RADIUS
        );
      }
      return;
    }

    if (!me.parachute_altitude_reached) {
      console.log("Altitude", me.getAltitudeAbs());
      if (Math.abs(me.getAltitudeAbs() - PARACHUTE_ALTITUDE) <=
          EPSILON_ALTITUDE) {
        me.parachute_altitude_reached = true;
      }
      return;
    }

    if (!me.parachute_yaw_reached) {
      console.log("Yaw:", me.getYaw());
      if (Math.abs(me.getYaw() - PARACHUTE_YAW) <= YAW_EPSILON) {
        me.parachute_yaw_reached = true;
        me.setTargetCoordinates(
          PARACHUTE_POINT.latitude,
          PARACHUTE_POINT.longitude,
          PARACHUTE_POINT.altitude
        );
      }
      return;
    }

    if (pointReached(me, PARACHUTE_POINT, PARACHUTE_EPSILON)) {
      deployParachute(me);
    }
  };

  me.onGetMsg = function (msg) {
    var msgDict = JSON.parse(msg);

    if (msgDict.hasOwnProperty("status")) {
      switch (msgDict.status) {

      case "running":
        me.started = true;
        if (me.id === leaderId(me)) {
          me.sendMsg(JSON.stringify(
            {id: me.id, inAir: true, state: "Flying", type: "state"}
          ));
        }
        break;

      case "stopped":
        if (me.id === me.id_list[0]) {
          me.landing = true;
          me.going_to_rally = true;
          me.loiter(
            RALLY_POINT.latitude,
            RALLY_POINT.longitude,
            me.flightAltitude,
            LOITER_RADIUS
          );
        }
        break;

      default:
        console.log("Unknown status:", me.msgDict.status);
      }
      return;
    }

    switch (msgDict.type) {

    case "state":
      if (msgDict.state === "Landed") {
        me.id_list.splice(me.id_list.indexOf(msgDict.id), 1);
        if (me.id === me.id_list[0]) {
          me.landing = true;
          me.going_to_rally = true;
          me.loiter(
            RALLY_POINT.latitude,
            RALLY_POINT.longitude,
            me.flightAltitude,
            LOITER_RADIUS
          );
        }
      }
      break;

    default:
      console.log("Unknown message type: " + msgDict.type);
    }
  };
}(console, me));
